from typing import Tuple, NamedTuple
import itertools
import torch.nn as nn
import numpy as np

from protein_design.constants import BLOSUM


class MLP(nn.Module):
    def __init__(self, **kwargs):
        super(MLP, self).__init__()

        self.data_dim = kwargs["data_dim"]
        self.hid_dim = kwargs["hid_dim"]
        self.loss_fn = nn.MSELoss()

        self.model = nn.Sequential(
            nn.Linear(self.data_dim, self.hid_dim),
            nn.ELU(),
            nn.Linear(self.hid_dim, 1),
            nn.Tanh(),
        )

    def forward(self, X):
        return self.model(X).squeeze()

    def loss(self, X, y):
        y_pred = self.model(X)
        return self.loss_fn(y, y_pred)


class GPParams(NamedTuple):
    homo_noise: float
    beta: float
    c: float
    d: float


class SequenceGP:
    def __init__(self, homo_noise=0.1, beta=0.1, c=1, d=2):
        self.params = GPParams(homo_noise, beta, c, d)
        self.K = None

    def kernel(self, Xi, Xj) -> float:
        """Return similarity of two amino acid sequences

        Parameters
        ----------
        Xi
            Integer encoded amino acid sequence
        Xj
            Integer encoded amino acid sequence

        Returns
        -------
            K(Xi, Xj) (measure of similarity between Xi and Xj)
        """
        beta = self.params.beta
        c = self.params.c
        d = self.params.d
        kij = np.prod(BLOSUM[(Xi, Xj)] ** beta)
        kii = np.prod(BLOSUM[(Xi, Xi)] ** beta)
        kjj = np.prod(BLOSUM[(Xj, Xj)] ** beta)
        k = kij / (np.sqrt(kii * kjj))
        k = np.exp(c * k)
        k = (k + c) / d
        return k

    def _fill_K(self):
        self.K = np.zeros((self.N, self.N))
        homo_noise = self.params.homo_noise
        for i in range(self.N):
            for j in range(i, self.N):
                kij = self.kernel(self.X[i], self.X[j])
                if i == j:
                    kij += homo_noise
                self.K[i, j] = kij
                self.K[j, i] = kij

    def fit(self, X: np.ndarray, y: np.ndarray) -> None:
        """Fit the Gaussian Process model to data

        Parameters
        ----------
        X
            Array representing amino acid sequences. Each entry is an integer-encoded sequence
        y
            Scalar values to predict, one per sequence
        """
        self.X = X
        self.y = y.reshape(-1, 1)
        self.N = len(X)
        self._fill_K()

    def predict(self, Xstar: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Predict scalar property values for sequences Xstar

        Parameters
        ----------
        Xstar
            Array representing amino acid sequences. Each entry is an integer-encoded sequence

        Returns
        -------
            Tuple: mean and variance of prediction
        """
        M = len(Xstar)
        Kstar = np.zeros((M, self.N))
        for i, j in itertools.product(range(M), range(self.N)):
            kij = self.kernel(Xstar[i], self.X[j])
            Kstar[i, j] = kij
        Kstarstar = np.zeros((M, M))
        for i, j in itertools.product(range(M), range(M)):
            kij = self.kernel(Xstar[i], Xstar[j])
            Kstarstar[i, j] = kij
        Kinv = np.linalg.inv(self.K)
        mu_star = np.matmul(Kstar, np.matmul(Kinv, self.y))
        sigma_star = Kstarstar - np.linalg.multi_dot([Kstar, Kinv, Kstar.T])
        return mu_star.squeeze(), np.sqrt(np.diag(sigma_star)).squeeze()



from typing import Callable
from logging import Logger

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from tqdm import tqdm
import pandas as pd
import numpy as np
import plotly.express as px

from protein_design.constants import device
from protein_design.data import ProteinData, cycle, to_numpy
from protein_design.evaluator import regression_metrics


logger = Logger("protein_design")


def training_step(model: nn.Module, data_generator: Callable, optimizer, scheduler):
    X, y = next(data_generator)

    optimizer.zero_grad()
    loss = model.loss(X, y)

    loss.backward()
    optimizer.step()
    scheduler.step()

    return loss.item()


def validation_step(model: nn.Module, data_generator: Callable):
    X, y = next(data_generator)

    with torch.no_grad():
        loss = model.loss(X, y)

    return loss.item()


def train(
    model,
    X_train,
    X_test,
    save_fname,
    y_train=None,
    y_test=None,
    steps=10000,
    pbar_increment=100,
    batch_size=32,
    optimizer=None,
    scheduler=None,
):
    if optimizer is None:
        optimizer = torch.optim.Adam(model.parameters(), lr=1e-4)
    if scheduler is None:
        scheduler = torch.optim.lr_scheduler.CyclicLR(
            optimizer, 1e-4, 1e-3, cycle_momentum=False
        )

    if y_train is None:
        train_dataset = ProteinData(X_train, X_train)
        test_dataset = ProteinData(X_test, X_test)
    else:
        train_dataset = ProteinData(X_train, y_train)
        test_dataset = ProteinData(X_test, y_test)

    input_shape = list(X_train.shape)
    input_shape[0] = batch_size
    logger.info(model)
    logger.info(f"Training on {len(train_dataset)} examples")
    logger.info(f"Testing on {len(test_dataset)} examples")
    logger.info(f"Model placed on {device}")
    model.double()
    model = model.to(device)

    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True)

    train_cycle = cycle(train_loader)
    test_cycle = cycle(test_loader)

    train_losses, val_losses = [], []

    best_model = None

    best_val_loss = 1e9
    progress_format = "train loss: {:.6f} val loss: {:.6f}"
    with tqdm(total=steps, desc=progress_format.format(0, 0)) as pbar:
        for i in range(steps):
            train_loss = training_step(model, train_cycle, optimizer, scheduler)
            val_loss = validation_step(model, test_cycle)
            train_losses.append(train_loss)
            val_losses.append(val_loss)
            if val_loss < best_val_loss:
                best_val_loss = val_loss
                torch.save(model.state_dict(), save_fname)
                best_model = model
            if (i + 1) % pbar_increment == 0:
                pbar.set_description(progress_format.format(train_loss, val_loss))
                pbar.update(pbar_increment)

    df_result = pd.DataFrame()
    df_result["steps"] = list(range(steps)) + list(range(steps))
    df_result["loss"] = train_losses + val_losses
    df_result["stage"] = ["train"] * steps + ["val"] * steps
    fig = px.line(df_result, x="steps", y="loss", color="stage")
    fig.show()

    if y_train is not None:
        y_true, y_pred = [], []
        for x, y in test_loader:
            y = to_numpy(y)
            y_true.extend(list(y))
            pred = model(x)
            pred = to_numpy(pred).squeeze()
            y_pred.extend(list(pred))
        regression_metrics(np.array(y_true), np.array(y_pred), plot=True)

    return best_model



from random import shuffle


def random_split(X, y=None, frac=0.8):
    """Random split of data

    Parameters
    ----------
    X
        Inputs
    y, optional
        Outputs, by default None
    frac, optional
        Fraction of data to use as training data, by default 0.8
    """
    idx = list(range(len(X)))

    if y is not None:
        assert len(y) == len(X)

    shuffle(idx)
    n_train = int(frac * len(X))

    idx_train, idx_test = idx[:n_train], idx[n_train:]

    if y is not None:
        return X[idx_train], y[idx_train], X[idx_test], y[idx_test]
    else:
        return X[idx_train], X[idx_test]



import numpy as np
import torch
from torch.utils.data import Dataset

from protein_design.constants import device


def to_tensor(X) -> torch.tensor:
    return X.to(device) if torch.is_tensor(X) else torch.from_numpy(X).to(device)


def to_numpy(X) -> np.ndarray:
    return X.detach().cpu().numpy() if torch.is_tensor(X) else np.array(X)


class ProteinData(Dataset):
    def __init__(self, X, y):
        self.X = to_tensor(X)
        self.y = to_tensor(y)

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        return self.X[idx], self.y[idx]


def cycle(dataloader):
    while True:
        for data in dataloader:
            yield data



import itertools
from pathlib import Path
from typing import List, Optional, Tuple
from logging import Logger

from Bio import SeqIO
import numpy as np
from tqdm import tqdm

from protein_design.constants import AA, AA_IDX, IDX_AA


logger = Logger("protein_design")


def read_fasta(fname: Path, format: str = "fasta") -> List[str]:
    """Read a file of sequences and returns a list of sequences

    Parameters
    ----------
    fname : Path
        Path to sequence file
    format : str, optional
        Format of the sequence file, by default "fasta"

    Returns
    -------
    List[str]
        List of sequences
    """
    return [record.seq for record in SeqIO.parse(fname, format)]


def write_fasta(
    fname: Path, seqs: List[str], headers: Optional[List[str]] = None
) -> None:
    """Write sequences to a fasta file

    Parameters
    ----------
    fname
        Full path of file name to save as
    seqs
        List of sequences
    headers, optional
        List of headers, by default None, in which case sequences will be labeled as
        integers starting from 0
    """
    if headers is None:
        headers = [f">{i}" for i in range(len(seqs))]
    records = []
    for i, seq in enumerate(seqs):
        records.append(headers[i])
        records.append(seq)
    with open(fname, "w") as fp:
        fp.write("\n".join(records))


def seq_to_onehot(seq: str, max_len: Optional[int] = None) -> np.ndarray:
    """Convert a sequence to a 2D one-hot encoded matrix

    Parameters
    ----------
    seq : str
        Single letter amino acid sequence, possibly containing gap '-' characters
    max_len : Optional[int], optional
        If provided, the returned one-hot encoded matrix will have shape (max_len, 21)
        Sequences longer than max_len are truncated at the tail (C-terminus)
        Gap characters are added to sequences shorter than max_len at the tail (C-terminus), by default None

    Returns
    -------
    np.ndarray
        Array of shape (length of seq, 21) or (max_len, 21) if max_len is provided.
    """
    seq = seq.upper()

    if max_len is None:
        max_len = len(seq)
    elif max_len > len(seq):
        seq = seq + "-" * (max_len - len(seq))
    elif max_len < len(seq):
        seq = seq[:max_len]

    onehot = np.zeros((max_len, len(AA)))
    for i in range(max_len):
        onehot[i, AA_IDX[seq[i]]] = 1

    return onehot


def seqs_to_onehot(seqs: List[str], flatten: bool = False) -> np.ndarray:
    """Converts a list of amino acid sequences to a 3D one-hot encoding matrix

    Parameters
    ----------
    seqs : list[str]
        List of amino acid sequences
    flatten : bool
        If True, flattens each one-hot encoded sequence

    Returns
    -------
    np.ndarray
        Array of shape (numer of sequences, maximum length of sequences, 21)
    """
    max_len = max(len(seq) for seq in seqs)

    onehots = []
    for seq in seqs:
        onehots.append(seq_to_onehot(seq, max_len=max_len))

    if flatten:
        result = np.array(onehots)
        N, L, D = result.shape
        return result.reshape(N, L * D)
    else:
        return np.array(onehots)


def integer_to_seqs(mat: np.ndarray) -> List[str]:
    """Convert an integer encoded list of sequences back into letters

    Parameters
    ----------
    mat
        Integer encoded sequences

    Returns
    -------
        List of amino acid strings
    """
    seqs = []
    for row in mat:
        seq = "".join(IDX_AA[i] for i in row)
        seqs.append(seq)
    return seqs


def seqs_to_integer(seqs: List[str]) -> np.ndarray:
    """Converts a list of amino acid sequences to a 2D array of integer-encoded sequences

    Parameters
    ----------
    seqs
        List of amino acid sequences

    Returns
    -------
        Array of integer-encoded sequences
    """
    onehot = seqs_to_onehot(seqs)
    return np.argmax(onehot, axis=-1)


def probs_to_seqs(mat: np.ndarray, sample: bool = True) -> List[str]:
    """Convert a probability distribution over sequences, marginalized per position,
    into a list of sequences

    Parameters
    ----------
    mat
        Probability matrix of sequences, shape B x L x D where
            B is the number of sequences,
            L is the sequence length and
            D is the number of letters to sample from (typically 20 amino acids + 1 gap)
    sample
        If True (default), samples from the probability distribution per position. If False,
        just takes the most likely amino acid at each position.

    Returns
    -------
        List of amino acid strings
    """
    B, L, D = mat.shape
    int_seqs = np.zeros((B, L))
    for b, l in itertools.product(range(B), range(L)):
        p = mat[b, l]
        if sample:
            int_seqs[b, l] = np.random.choice(D, p=p)
        else:
            int_seqs[b, l] = np.argmax(p)
    return integer_to_seqs(int_seqs)


def trim_gaps(mat: np.ndarray, perc: float = 0.8) -> np.ndarray:
    """Removes the columns of an input one-hot sequence array which more than {cutoff}
    entries have gaps

    Parameters
    ----------
    mat
        One-hot sequence array, shape B x L x D
    perc
        Percentage of sequences above which a position is to be trimmed, by default 0.8

    Returns
    -------
        One-hot sequence array with gaps trimmed
    """
    B = mat.shape[0]
    frac = np.sum(mat, axis=0) / B  # shape (L, D)
    mask = frac[:, AA_IDX["-"]] < perc  # shape (L,)
    trimmed = mat[:, mask, :]
    return trimmed



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from protein_design.data import to_tensor
from protein_design.constants import AA, AA_IDX, device
from protein_design.sequence import integer_to_seqs, seqs_to_integer


def KLLoss(mean, logvar):
    return torch.mean(
        -0.5 * torch.sum(1 + logvar - mean ** 2 - logvar.exp(), dim=1), dim=0
    )


class VAE(nn.Module):
    def __init__(self, **kwargs):
        super(VAE, self).__init__()

        self.seqlen = kwargs["seqlen"]
        self.n_tokens = kwargs["n_tokens"]
        self.latent_dim = kwargs["latent_dim"]
        self.enc_units = kwargs["enc_units"]
        self.kl_weight = kwargs["kl_weight"]

        self.encoder = nn.Sequential(
            nn.Linear(self.seqlen * self.n_tokens, self.enc_units),
            nn.ELU(),
        )
        self.mean = nn.Linear(self.enc_units, self.latent_dim)
        self.var = nn.Linear(self.enc_units, self.latent_dim)

        self.decoder = nn.Sequential(
            nn.Linear(self.latent_dim, self.enc_units),
            nn.ELU(),
            nn.Linear(self.enc_units, self.seqlen * self.n_tokens),
        )
        self.getprobs = nn.Softmax(dim=-1)

        self.recon_loss_fn = nn.MSELoss()
        self.kl_loss_fn = KLLoss

    def encode(self, x):
        z = self.encoder(x)
        mean = self.mean(z)
        logvar = self.var(z)
        return [mean, logvar]

    def decode(self, z):
        xhat = self.decoder(z).view(-1, self.seqlen, self.n_tokens)
        xhat = self.getprobs(xhat)
        return xhat

    def reparameterize(self, mean, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mean + eps * std

    def forward(self, x):
        mean, logvar = self.encode(x)
        z = self.reparameterize(mean, logvar)
        return [self.decode(z), x, mean, logvar]

    def loss(self, x, x_target):
        xhat, _, mean, logvar = self.forward(x)

        x_target = x_target.view(-1, self.seqlen, self.n_tokens)
        # x = torch.argmax(x, -1).flatten()
        # x_target = x_target.flatten(end_dim=1)
        # recon_loss = F.cross_entropy(x_target, x.type(torch.long))
        recon_loss = self.recon_loss_fn(xhat, x_target)
        kl_loss = self.kl_loss_fn(mean, logvar)

        loss = recon_loss + self.kl_weight * kl_loss

        return loss

    def sample(self, num_samples):
        z = (
            torch.randn(num_samples, self.latent_dim)
            .type(torch.DoubleTensor)
            .to(device)
        )
        return self.decode(z).cpu().detach().numpy()

    def reconstruct(self, x):
        recon = self.forward(x)
        return recon[0]


class Attention(nn.Module):
    def __init__(self, **kwargs):
        super(Attention, self).__init__()

        self.n_head = kwargs["n_head"]
        self.d_model = kwargs["d_model"]
        self.d_k = kwargs["d_k"]
        self.d_v = kwargs["d_v"]

        self.W_Q = nn.Linear(self.d_model, self.n_head * self.d_k)
        self.W_K = nn.Linear(self.d_model, self.n_head * self.d_k)
        self.W_V = nn.Linear(self.d_model, self.n_head * self.d_v)
        self.W_O = nn.Linear(self.n_head * self.d_v, self.d_model)

        self.layer_norm = nn.LayerNorm(self.d_model)
        self.dropout = nn.Dropout(kwargs["dropout"])

    def forward(self, q, k, v):
        batch, len_q, _ = q.size()
        batch, len_k, _ = k.size()
        batch, len_v, _ = v.size()

        Q = self.W_Q(q).view([batch, len_q, self.n_head, self.d_k])
        K = self.W_K(k).view([batch, len_k, self.n_head, self.d_k])
        V = self.W_V(v).view([batch, len_v, self.n_head, self.d_v])

        Q = Q.transpose(1, 2)
        K = K.transpose(1, 2).transpose(2, 3)
        V = V.transpose(1, 2)

        # after transpose:
        # Q: (batch, n_head, len_q, d_k)
        # K: (batch, n_head, d_k, len_q)
        # V: (batch, n_head, len_q, d_v)

        attention = torch.matmul(Q, K)
        # attention: (batch, n_head, len_q, len_q)

        attention = attention / np.sqrt(self.d_k)

        attention = F.softmax(attention, dim=-1)

        output = torch.matmul(attention, V)
        # output: (batch, n_head, len_q, d_v)

        output = output.transpose(1, 2).reshape([batch, len_q, self.d_v * self.n_head])
        # output: (batch, len_q, d_v * n_head)

        output = self.W_O(output)

        output = self.dropout(output)

        output = self.layer_norm(output + q)

        return output


def label_smoothed_nll_loss(lprobs, target, epsilon: float = 1e-8, ignore_index=None):
    """Adapted from fairseq

    Parameters
    ----------
    lprobs
        Log probabilities of amino acids per position
    target
        Target amino acids encoded as integer indices
    epsilon
        Smoothing factor between 0 and 1, by default 1e-8
    ignore_index, optional
        Amino acid (encoded as integer) to ignore, by default None

    Returns
    -------
        Negative log-likelihood loss
    """
    nll_loss = -lprobs.gather(dim=-1, index=target)
    smooth_loss = -lprobs.sum(dim=-1, keepdim=True)
    if ignore_index is not None:
        pad_mask = target.eq(ignore_index)
        nll_loss.masked_fill_(pad_mask, 0.0)
        smooth_loss.masked_fill_(pad_mask, 0.0)
    else:
        nll_loss = nll_loss.squeeze(-1)
        smooth_loss = smooth_loss.squeeze(-1)
    nll_loss = nll_loss.sum()
    smooth_loss = smooth_loss.sum()
    eps_i = epsilon / lprobs.size(-1)
    loss = (1.0 - epsilon) * nll_loss + eps_i * smooth_loss
    return loss


def positional_embedding(d_model: int, length: int) -> torch.tensor:
    result = -torch.exp(torch.arange(0, d_model, 2) * (np.log(10000) / d_model))

    numbers = torch.arange(0, length)
    numbers = numbers.unsqueeze(0)
    numbers = numbers.unsqueeze(2)

    result = numbers * result
    result = torch.cat((torch.sin(result), torch.cos(result)), 2)
    result = to_tensor(result)

    return nn.Parameter(result, requires_grad=True)


class BERT(nn.Module):
    def __init__(self, **kwargs):
        super(BERT, self).__init__()

        self.d_model = kwargs["d_model"]

        self.embed = nn.Embedding(len(AA) + 1, self.d_model)
        self.positional_embedding = None

        self.layer1 = Attention(**kwargs)
        self.layer2 = Attention(**kwargs)
        self.linear = nn.Linear(self.d_model, len(AA))

        self.num_mask = kwargs["num_mask"]

    def forward(self, x):
        embeds = self.embed(x)
        if self.positional_embedding is None:
            self.positional_embedding = positional_embedding(self.d_model, x.shape[1])
        embeds += self.positional_embedding

        h1 = self.layer1(embeds, embeds, embeds)
        h2 = self.layer2(h1, h1, h1)

        B, L, D = h2.shape
        logits = self.linear(h2.view(B * L, D)).view(B, L, len(AA))
        lprobs = F.log_softmax(logits, dim=-1)

        return lprobs

    def loss(self, x, _):
        B, L = x.shape
        start_idx = np.random.choice(range(L - self.num_mask))
        masked_idx = tuple(range(start_idx, start_idx + self.num_mask))

        x_target = x[:, masked_idx]
        x[:, masked_idx] = 21

        lprobs = self.forward(x)[:, masked_idx]
        loss = label_smoothed_nll_loss(
            lprobs.view(-1, lprobs.size(-1)), x_target.view(-1, 1), ignore_index=20
        )
        return loss / (B * self.num_mask)

    def sample(self, seq, n=1000, rm_aa=""):
        masked = [i for i in range(len(seq)) if seq[i] == "X"]
        x = seqs_to_integer([seq])
        x = [aa if i not in masked else 21 for i, aa in enumerate(x[0])]
        x = np.expand_dims(np.array(x), 0)
        x = torch.from_numpy(x).type(torch.LongTensor)

        with torch.no_grad():
            lprobs = self.forward(x).cpu().detach().numpy()

        def _sample(i) -> int:
            probs = np.exp(lprobs[i])
            probs[20] = 0
            probs[21] = 0
            for aa in rm_aa.split(","):
                probs[AA_IDX[aa.upper()]] = 0
            return np.random.choice(22, p=probs)

        x = seqs_to_integer([seq])
        all_sampled = []
        for _ in range(n):
            sampled = []
            for i in range(len(seq)):
                if i in masked:
                    sampled.append(_sample(i))
                else:
                    sampled.append(x[i])
            all_sampled.append(sampled)

        return integer_to_seqs(np.array(all_sampled))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['protein_design']

package_data = \
{'': ['*']}

install_requires = \
['Sphinx>=4.4.0,<5.0.0',
 'biopython>=1.79,<2.0',
 'black>=22.1.0,<23.0.0',
 'invoke>=1.6.0,<2.0.0',
 'numpy==1.21.0',
 'pandas==1.1.5',
 'plotly>=5.6.0,<6.0.0',
 'py3Dmol>=1.8.0,<2.0.0',
 'scipy==1.6.1',
 'seaborn>=0.11.2,<0.12.0',
 'sklearn>=0.0,<0.1',
 'torch>=1.10.2,<2.0.0',
 'tqdm>=4.62.3,<5.0.0',
 'umap-learn>=0.5.2,<0.6.0']

setup_kwargs = {
    'name': 'protein-design',
    'version': '0.4.2',
    'description': 'Python tools for protein design',
    'long_description': None,
    'author': 'Tianyu Lu',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



from typing import Dict

import numpy as np
import pandas as pd
from scipy.stats import spearmanr, pearsonr
from sklearn.metrics import ndcg_score
import plotly.express as px


def regression_metrics(y_true, y_pred, plot=False, hover_data=None) -> Dict[str, float]:
    """Compute common metrics used to evalute regression models

    Parameters
    ----------
    y_true
        Actual values
    y_pred
        Predicted values
    plot
        If True, plots a scatterplot of true vs. predicted values, by default False
    hover_data
        If provided, includes this metadata as plotly hover_data

    Returns
    -------
        Dictionary mapping regression metric to value
    """
    spearman = spearmanr(y_true, y_pred)[0]
    pearson = pearsonr(y_true, y_pred)[0].item()
    mae = np.mean(np.abs(y_true - y_pred))
    idx = np.argsort(-y_true)
    rank = np.argsort(idx)
    ndcg = ndcg_score(rank.reshape(1, -1), -y_pred.reshape(1, -1))

    if plot:
        title = f"Spearman: {spearman:.3f} Pearson: {pearson:.3f} MAE: {mae:.3f} NDCG: {ndcg:.3f}"
        df = pd.DataFrame()
        df["y_true"] = y_true
        df["y_pred"] = y_pred
        df["hover"] = hover_data
        fig = px.scatter(df, x="y_true", y="y_pred", title=title, hover_data=["hover"])
        fig.show()

    return {"spearman": spearman, "pearson": pearson, "mae": mae, "ndcg": ndcg}



from typing import NewType, List, Optional, Union, Tuple
from pathlib import Path
import subprocess
import itertools

import numpy as np
import py3Dmol
import pandas as pd
from tqdm import tqdm
from Bio.SeqUtils import seq1
from Bio.PDB import PDBParser, PDBIO, Selection, Structure, Chain, Residue
from Bio.PDB.PDBExceptions import PDBConstructionException


parser = PDBParser()
io = PDBIO()
Resid = NewType("Resid", Tuple[int, str, Tuple[str, int, str]])


def download_pdb(pdb_code: str) -> Structure:
    """Download a pdb file from RCSB and load it as a Biopython Structure

    Parameters
    ----------
    pdb_code
        PDB code

    Returns
    -------
        Bio.PDB.Structure
    """
    pdb_code = pdb_code.upper()
    subprocess.run(
        ["wget", f"https://files.rcsb.org/download/{pdb_code}.pdb"], check=True
    )
    structure = parser.get_structure("s", f"{pdb_code}.pdb")
    return structure


def remove_hetero(structure: Structure, subset: Optional[str] = None) -> None:
    """Removes all water molecules in {structure}

    Parameters
    ----------
    structure
        Bio.PDB.Structure
    subset
        Comma-delimited string of the residues to remove, e.g. "HOH,SO4" to remove all waters and SO4 molecules
    """
    residues_to_remove = []
    for residue in Selection.unfold_entities(structure, "R"):
        resname = residue.get_resname()
        if subset is not None:
            if resname in subset.split(","):
                residues_to_remove.append(residue)
        else:
            if seq1(resname) == "X":
                residues_to_remove.append(residue)
    for r in residues_to_remove:
        r.get_parent().detach_child(r.get_id())


def get_xyz(entity: List) -> np.ndarray:
    """Get xyz coordinates of all atoms in the entity list

    Parameters
    ----------
    entity
        Any one of Bio.PDB.(Structure, Model, Chain, Residue, Atom)

    Returns
    -------
        N x 3 numpy array of xyz coordinates, where N is the number of atoms
    """
    coords = []
    for atom in Selection.unfold_entities(entity, "A"):
        coords.append(atom.coord)
    return np.array(coords)


def get_res_data(residue: Residue) -> Resid:
    """Get the model, chain and residue id of the input {residue}

    Parameters
    ----------
    residue
        Bio.PDB.Residue

    Returns
    -------
    Resid tuple containing three elements:
        model:
            Model number
        chain:
            Single letter chain
        resid:
            Full residue id: (heteroatom flag, numerical id, insertion code)
    """
    _, model, chain, full_resid = residue.get_full_id()
    return (model, chain, full_resid)


def get_interface(c1: Chain, c2: Chain, cutoff: float = 5.0) -> List[Resid]:
    """Given two chains, find the residues where at least one pair of atoms is
    within {cutoff} Angstroms of each other

    Parameters
    ----------
    c1
        Chain 1
    c2
        Chain 2
    cutoff, optional
        Interatomic distance in Angstroms to consider a pair of residues as
        "in contact" and thus belonging to the interface, by default 5.0

    Returns
    -------
        List of residue identifiers used to subset the structure containing chains 1 and 2
    """
    res1_interface = []
    res2_interface = []
    for res1, res2 in itertools.product(c1.get_residues(), c2.get_residues()):
        closest_dist = 1e6
        res1_coords, res2_coords = get_xyz(res1), get_xyz(res2)
        for coord in res1_coords:
            dists = np.linalg.norm(coord.reshape(1, 3) - res2_coords, axis=1)
            curr_dist = np.min(dists)
            if curr_dist < closest_dist:
                closest_dist = curr_dist
        if closest_dist <= cutoff:
            res1_interface.append(get_res_data(res1))
            res2_interface.append(get_res_data(res2))
    return list(set(res1_interface).union(set(res2_interface)))


def subset_structure(structure: Structure, subset_list: List[Resid]) -> None:
    """Subsets the {structure} to keep only those residues in {subset_list}.
    Note: this function modifies the input {structure} in place.

    Parameters
    ----------
    structure
        Bio.PDB.Structure
    subset_list
        List of residue identifiers to keep, obtained from e.g. get_interface()
    """
    to_remove = []
    for model, chain, full_resid in subset_list:
        for residue in Selection.unfold_entities(structure[model][chain]):
            if residue.get_full_id() != full_resid:
                to_remove.append(residue)
    for residue in to_remove:
        residue.get_parent().detach_child(residue)


def show_from_pdbid(pdbid: str) -> None:
    """Show an interactive structure of {pdbid} using py3Dmol

    Parameters
    ----------
    pdbid
        PDB code
    """
    view = py3Dmol.view(query=f"pdb:{pdbid}", options={"doAssembly": True})
    view.setStyle({"cartoon": {"color": "spectrum"}})
    view.show()


def show_from_file(pdb_file: Union[Path, str], chains: str = "A") -> None:
    """Show an interactive structure from chains {chains} from file {pdb_file}

    Parameters
    ----------
    pdb_file
        Path to pdb file
    chains
        String of chains to show, by default "A"
    """
    with open(pdb_file) as fp:
        interface_pdb = "".join(
            line
            for line in fp.readlines()
            if line.startswith("ATOM") and line[21] in chains
        )

    view = py3Dmol.view()
    view.addModelsAsFrames(interface_pdb)
    for chain in chains:
        view.setStyle({"chain": chain}, {"cartoon": {"color": "spectrum"}, "stick": {}})
    view.show()

    view.setHoverable(
        {},
        True,
        """function(atom,viewer,event,container) {
                    if(!atom.label) {
                        atom.label = viewer.addLabel(atom.resn+":"+atom.atom,{position: atom, backgroundColor: 'mintcream', fontColor:'black'});
                    }}""",
        """function(atom,viewer) { 
                    if(atom.label) {
                        viewer.removeLabel(atom.label);
                        delete atom.label;
                    }
                    }""",
    )


def structure_to_sequence(structure: Structure) -> str:
    """Convert a Bio.PDB.Structure into a sequence.

    Parameters
    ----------
    structure
        Bio.PDB.Structure

    Returns
    -------
        Sequence
    """
    seq = []
    for residue in Selection.unfold_entities(structure, "R"):
        aa = seq1(residue.get_resname())
        seq.append(aa)
    return "".join(seq)


if __name__ == "__main__":
    entries = []
    data_dir = Path("data/sabdab/2022-04-03/")
    summary_fp = data_dir / "sabdab_nano_summary_all.tsv"
    df = pd.read_csv(summary_fp, sep="\t")
    df = df.loc[df["pdb"] != "6iyn"]
    df.dropna(subset=["Hchain"], inplace=True)
    pdb_codes = df["pdb"].to_list()
    nanobody_chains = df["Hchain"].to_list()
    structures_dir = data_dir / "imgt"
    for pdb, chain in tqdm(zip(pdb_codes, nanobody_chains)):
        pdb_fp = structures_dir / f"{pdb}.pdb"
        structure = parser.get_structure(pdb, pdb_fp)
        remove_hetero(structure)
        try:
            seq = structure_to_sequence(structure[0][chain])
            entries.append(f">{pdb}_{chain}\n{seq}")
        except KeyError:
            print(pdb, chain)
            continue
        except PDBConstructionException:
            print(pdb, chain)
            continue
    with open(data_dir / "vhh.fasta", "w") as fp:
        fp.write("\n".join(entries))
